#include "precompile.h"
#include "solver.h"

void ReadInput(Input* input) {
	FILE* in_file = fopen("recycle.in", "rt");
	fscanf(in_file, "%d%d%d%d", &input->N, &input->M, &input->P, &input->C); 
	input->board.resize(input->N);
	for (int row = 0; row < input->N; ++row) {
		input->board[row].resize(input->M);
		for (int col = 0; col < input->M; ++col) {
			fscanf(in_file, "%d", &input->board[row][col]);
		}
	}
}

void PrintOutput(const Solution& output) {
	FILE* out_file = fopen("recycle.out", "wt");
	fprintf(out_file, "%d\n", (int)output.operations.size());
	for (size_t oper_ind = 0; oper_ind < output.operations.size(); ++oper_ind) {
		const Operation& operation = output.operations[oper_ind];
		fprintf(out_file, "%c %d %d\n", (operation.type == ROW_SHIFT ? 'L' : 'U'), operation.index + 1, operation.times);
	}
}

Solver solver;

void Improve(const Input& input, Solution* sol) {
	Solution next(*sol);
	int curr_score = GetScore(input, *sol);
	int improvements = 0;

	clock_t beg = clock();
	while (true) {
		if (clock() - beg > TIME_TO_IMPROVE) {
			break;
		}

		random_shuffle(next.operations.begin(), next.operations.end());
		int next_score = GetScore(input, next);
		if (next_score > curr_score) {
			curr_score = next_score;
			*sol = next;
			++improvements;
		}
	}
#ifdef TEST
	printf("Improvement count: %d\n", improvements);
#endif
}

void ModifyInput(Input& input, const Solution& s) {
	Board board(input.board);
	
	vector<int> temp;
	for (size_t oper_ind = 0; oper_ind < s.operations.size(); ++oper_ind) {
		const Operation& operation(s.operations[oper_ind]);
		if (operation.type == ROW_SHIFT) {
			int row_ind = operation.index;
			temp = board[row_ind];
			for (int col_index = 0; col_index < input.M; ++col_index) {
				size_t curr_col = (col_index + operation.times) % input.M;
				board[row_ind][col_index] = temp[curr_col];
			}
		}
		else {
			int col_ind = operation.index;
			temp.resize(input.N);
			for (int r_ind = 0; r_ind < input.N; ++r_ind) {
				temp[r_ind] = board[r_ind][col_ind];
			}

			for (int row_index = 0; row_index < input.N; ++row_index) {
				size_t curr_row = (row_index + operation.times) % input.N;
				board[row_index][col_ind] = temp[curr_row];
			}
		}
	}

	input.board = board;
}

int main() {
//#ifdef TEST
	clock_t beg = clock();
//#endif

	Input input;
	ReadInput(&input);

	Input temp(input);
	Solution sol;

	solver.Solve(input, &sol);

	Improve(input, &sol);

	int iter_cnt = 0;
	while (iter_cnt + 1 < ITERATIONS) {
		++iter_cnt;
		//printf("%d\n", GetScore(temp, sol));
		input = temp;
		ModifyInput(input, sol);
		solver.Clear();
		Solution sol2;
		solver.Solve(input, &sol2);
		sol.operations.insert(sol.operations.end(), sol2.operations.begin(), sol2.operations.end());
	}
	//printf("%d\n", GetScore(temp, sol));
	PrintOutput(sol);
//#ifdef TEST
	printf("Total time: %.3lf\n", (clock() - beg) / (double)CLOCKS_PER_SEC);
//#endif
	return 0;
}




